/*
  
pfsgb.c - a scenegraph loader for file suffix "sgb"

usage:  pfNode *foo = pfdLoadFile("foo.sgb") ;


John Kelso, kelso@nist.gov
March, 2004
  
A pfGroup node is created- under it are loaded nodes and files

A sgb file has the following format:

# this is a comment
# keyword matching is case insensative

# The "active node" is the node under which other nodes are placed.
# Initially the top level group node is the active node.

# create a pfGroup, pfDCS or pfSCS node named "nodeName" under the active
# node, and make it the new active node.  An optional transformation can be
# given.  A node name can be left unspecified by using the keyword NULL

GROUP [ nodeName ]
DCS   [ nodeName [ x y z [ h p r [ s [ sy sz ] ] ] ] ] 
SCS   [ nodeName [ x y z [ h p r [ s [ sy sz ] ] ] ] ] 

# make the node named "nodeName" the active node- subsequent nodes and files
# will be placed under this node.  If no node is specified, use the top
# level node
UNDER [ nodeName ]

# load files named under the active node
# duplicate files are automatically instanced
FILE filename.ext [ filename.ext ... ]


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Performer/pf.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfSCS.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pfutil.h>
#include <Performer/pfdu.h>

#include "getString.C"
#include "normalize.C"
#include "parse.C"

#define MAX_STRLEN (2048)

// function templates
void cleanup(void) ;

pfMatrix getMatrix(int argc, char **argv) ;
char* getSuffix(char* fname
);
// file name and pointer to pfNode where the file is loaded
struct fileNode {
  char *name ;
  pfNode *node ;
} ;

// evil global variables

// input line number for error messages
int lineNum ;

// an array of unique fileNodes
fileNode *file ;

// how many of them?
int files ;

// group node to return ;
pfGroup *topgroup ;

// handy identity matrix
pfMatrix identity ;

// active node- might be a pfSCS or pfDCS cast to a pfGroup
pfGroup *active ;

extern "C" pfNode * pfdLoadFile_sgb (char *filename) {
  pfNotify(PFNFY_NOTICE, PFNFY_PRINT, "pfdLoadFile_sgb: %s", filename);
  FILE *fp = pfdOpenFile(filename);
  if (!fp)
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_sgb: unable to open %s for reading", filename);
  
  int notifyLevel = pfGetNotifyLevel() ;

  // in a DSO you need to initialize with assignment statements
  lineNum = 0 ;
  file = NULL ;
  files = 0 ;
  topgroup = new pfGroup() ;
  active = topgroup ;
  identity.makeIdent() ;

  // read in the file and store data
  char *line = NULL ;

  // loop through the input
  while (1) {
    if (notifyLevel >= PFNFY_INFO) {
      fprintf(stderr,"%8.8d   \r",lineNum) ;
      fflush(stderr) ;
    }
    lineNum++ ;

    if (getString(&line, fp)<0) {
      cleanup() ;
      break ;
    } else {
      if (normalize(line)) {
	int argc ;
	char **argv ;
	dpf_charsToArgcArgv(line, &argc, &argv) ;

	// how many characters to match
  	int l = strlen(argv[0]) ;
	if (l<3) l=3 ;
	// start parsing non-blank lines

	// DCS node
	if (!(strncasecmp(argv[0],"dcs",l))) {
	  pfMatrix mat ;
	  if (argc>=3)
	    mat = getMatrix(argc-2, &argv[2]) ;
	  else 
	    mat = identity ;
	  pfDCS *node = new pfDCS() ;
	  node->setMat(mat) ;
	  active->addChild(node) ;
	  active = (pfGroup *) node ;
	  if (argc>1 && (strcasecmp(argv[1],"NULL"))) {
	    if (!(node->setName(argv[1]))) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: line %d, "
		       "unable to name node \"%s\"", lineNum, argv[1]);
	  }

	// SCS node
	} else if (!(strncasecmp(argv[0],"scs",l))) {
	  pfMatrix mat ;
	  if (argc>=3)
	    mat = getMatrix(argc-2, &argv[2]) ;
	  else 
	    mat = identity ;
	  pfGroup *node;

	  if (mat == identity)
	    node = new pfGroup() ;
	  else
	    node = new pfSCS(mat) ;
	  active->addChild(node) ;
	  active = (pfGroup *) node ;
	  if (argc>1 && (strcasecmp(argv[1],"NULL"))) {
	    if (!(node->setName(argv[1]))) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: line %d, "
		       "unable to name node \"%s\"", lineNum, argv[1]);
	  }

	// group node
	} else if (!(strncasecmp(argv[0],"group",l))) {
	  pfGroup *node = new pfGroup() ;
	  active->addChild(node) ;
	  active = node;
	  if (argc>1 && (strcasecmp(argv[1],"NULL"))) {
	    if (!(node->setName(argv[1]))) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: line %d, "
		       "unable to name node \"%s\"", lineNum, argv[1]);
	  }

	  // set active node
	} else if (!(strncasecmp(argv[0],"under",l))) {
	  if (argc == 1) {
	    active = topgroup ;
	  } else if (argc == 2) {
	    pfNode *node = NULL ;
	    node = topgroup->lookup(argv[1], pfGroup::getClassType()) ;
	    if (!node)
	      node = topgroup->lookup(argv[1], pfDCS::getClassType()) ;
	    if (!node)
	      node = topgroup->lookup(argv[1], pfSCS::getClassType()) ;
	    if (!node) 
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: line %d, "
		       "unable to find node \"%s\"", lineNum, argv[1]);
	    else {
	      active = (pfGroup *) node ;
	    }
	  } else
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sgb: line %d, "
		     "missing node name", lineNum);

	  
	// finally, load files, instancing as we go
	} else if (!(strncasecmp(argv[0],"file",l))) {
	  // see if filename has already been loaded
	  int thisFile ;
	  for (thisFile=0; thisFile<files; thisFile++) {
	    if (!strcmp(file[thisFile].name,argv[1]))
	      break ;
	  }
	    
	  // didn't find it, so add it
	  if (thisFile>=files) { 
	    thisFile = files ;
	    files++ ;
	    file = (fileNode *) realloc(file, sizeof(fileNode)*files) ;
	    file[thisFile].name = strdup(argv[1]) ;
	    file[thisFile].node = pfdLoadFile(argv[1]) ;

	    // see if there is name
	    if (argc > 2)
	      file[thisFile].node->setName(argv[2]);
	    if (!file[thisFile].node)
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: line %d, "
		       "unable to load file %s", lineNum, argv[1]);
	  } //if (thisFile>=files)
	  if (file[thisFile].node != NULL)
	    active->addChild(file[thisFile].node) ;
	

	// finally, load files, instancing as we go
	} else if (!(strncasecmp(argv[0],"load",l))) {
	  // see if filename has already been loaded
	  int thisFile ;
	  for (thisFile=0; thisFile<files; thisFile++) {
	    if (!strcmp(file[thisFile].name,argv[1]))
	      break ;
	  }
	    
	  // didn't find it, so add it
	  if (thisFile>=files) { 
	    thisFile = files ;
	    files++ ;
	    file = (fileNode *) realloc(file, sizeof(fileNode)*files) ;
	    file[thisFile].name = strdup(argv[1]) ;
	    file[thisFile].node = pfdLoadFile(argv[1]) ;

	    if (!file[thisFile].node) {
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: line %d, "
		       "unable to load file %s", lineNum, argv[1]);
	    }
	    else {
	      // see if there is name
	      if (argc > 2)
		file[thisFile].node->setName(argv[2]);
	    }
	  } //if (thisFile>=files)
	  //if (!(strncasecmp(argv[0],"file",l)))
	  //  active->addChild(file[thisFile].node) ;


	// set active node
	} else if (!(strncasecmp(argv[0],"on",l))) {
	  // input is: ON nodename parentname
	  if (argc == 1) {
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sgb: "
		     "missing node name for ON statement.");
	  } else {
	    // get node by name
	    pfNode *node = NULL ;
	    node = topgroup->lookup(argv[1], pfNode::getClassType()) ;

	    if (argc == 3) {
	      // get parent node name
	      pfGroup* parentnode = NULL;
	      parentnode = (pfGroup*)topgroup->lookup(argv[2], pfGroup::getClassType()) ;
	      if (!parentnode)
		parentnode = (pfGroup*)topgroup->lookup(argv[2], pfDCS::getClassType()) ;
	      if (!parentnode)
		parentnode = (pfGroup*)topgroup->lookup(argv[2], pfSCS::getClassType()) ;
	      if (!parentnode) 
		pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: "
		       "unable to find parent node \"%s\"", argv[2]);
	      if (parentnode != NULL)
		active = parentnode ;
	    }
	    if (node != NULL)
	      active->addChild(node);
	    else
	      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: "
		       "unable to find node \"%s\"", argv[1]);
	  }
	} else {
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		       "pfdLoadFile_sgb: "
		       "no keyword matches: \"%s\"", argv[0]);
	}
      } //if (normalize(line))
    } //if (getString(&line, fp)<0)
  } // while (1)

  fclose(fp);
  return(topgroup) ;
}


void cleanup(void) {
}

pfMatrix getMatrix(int argc, char **argv) {
  pfCoord coord ;
  pfVec3  scale ;

  // initialize to identity
  coord.xyz = pfVec3(0.f, 0.f, 0.f) ;
  coord.hpr = pfVec3(0.f, 0.f, 0.f) ;
  scale = pfVec3(1.f, 1.f, 1.f) ;

  if (argc!=3 && argc!=6 && argc!=7 && argc!=9)
    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	     "pfdLoadFile_sgb: line %d, "
	     "invalid number of parameters", lineNum);

  if (argc >= 3) {
    if ((sscanf(argv[0],"%f",&coord.xyz[0])!=1) ||
	(sscanf(argv[1],"%f",&coord.xyz[1])!=1) ||
	(sscanf(argv[2],"%f",&coord.xyz[2])!=1))
      pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
	       "pfdLoadFile_sgb: line %d, "
	       "invalid number", lineNum);
    else if (argc >= 6) {
      if ((sscanf(argv[3],"%f",&coord.hpr[0])!=1) ||
	  (sscanf(argv[4],"%f",&coord.hpr[1])!=1) ||
	  (sscanf(argv[5],"%f",&coord.hpr[2])!=1))
	pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		 "pfdLoadFile_sgb: line %d, "
		 "invalid number", lineNum);
      
      else if (argc >= 7) {
	if ((sscanf(argv[6],"%f",&scale[0])!=1))
	  pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		   "pfdLoadFile_sgb: line %d, "
		   "invalid number", lineNum);
	
	else if (argc == 9) {
	  if ((sscanf(argv[7],"%f",&scale[1])!=1) ||
	      (sscanf(argv[8],"%f",&scale[2])!=1))
	    pfNotify(PFNFY_FATAL, PFNFY_PRINT, 
		     "pfdLoadFile_sgb: line %d, "
		     "invalid number", lineNum);
	} else {
	  scale[1] = scale[2] = scale[0] ;
	} // if (argc == 9)
      } // if (argc >= 7)
    } // if (argc >= 6)
  } // if (argc >= 3)

  pfMatrix matrix ;
  matrix.makeCoord(&coord) ;
  matrix.preScale(scale[0], scale[1], scale[2], matrix) ;

  return (matrix) ;
}

char* getSuffix(char* fname) {
  char *suf  = NULL;
  if (strlen(fname)>0) {
    suf = strrchr(fname, '.') ;
  }
  return suf;
}
